import http.client
import io
import json
import logging
import math
import os
import socket
import ssl
import time
import urllib
#from abc import abstractmethod
from typing import Callable, Optional, Union
from urllib.parse import parse_qs, urlparse

import numpy as np
import pytubefix
import pytubefix.exceptions
import requests
import urllib3
#from pytubefix.exceptions import VideoUnavailable, VideoPrivate, MaxRetriesExceeded
import yt_dlp
from PIL import Image
from pytubefix import YouTube
from pytubefix.innertube import _default_clients
from pytubefix.streams import Stream

from warp_beacon.jobs.download_job import DownloadJob
from warp_beacon.scraper.abstract import ScraperAbstract
from warp_beacon.scraper.exceptions import (BadProxy, TimeOut, Unavailable,
											extract_exception_message)
from warp_beacon.yt_auth import YtAuth

class YoutubeAbstract(ScraperAbstract):
	DOWNLOAD_DIR = "/tmp"
	YT_SESSION_FILE = '/var/warp_beacon/yt_session_%d.json'

	def __init__(self, account: tuple, proxy: dict=None) -> None:
		super().__init__(account, proxy)
		self._download_progress_threshold = 20

	def validate_session(self) -> int:
		try:
			logging.info("Validating YT session(s) ...")
			session_dir = os.path.dirname(self.YT_SESSION_FILE)
			for f in os.listdir(session_dir):
				if f.startswith("yt_session") and f.endswith(".json"):
					yt_sess_file = f"{session_dir}/{f}"
					if os.path.exists(yt_sess_file):
						account_index = int(f.split('_')[-1].rstrip('.json'))
						logging.info("Validating YT session #%d ...", account_index)
						yt_sess_data, exp = {}, 0
						with open(yt_sess_file, 'r', encoding="utf-8") as f:
							yt_sess_data = json.loads(f.read())
							exp = int(yt_sess_data.get("expires", 0))
						if exp <= time.time() + 60:
							yt_auth = YtAuth(account_index=account_index)
							requests_data = yt_auth.refresh_token(refresh_token=yt_sess_data.get("refresh_token", ""))
							if requests_data:
								yt_sess_data.update(requests_data)
								if yt_auth.safe_write_session(yt_sess_data):
									logging.info("YT session #%d validated", account_index)
		except Exception as e:
			logging.error("Failed to refresh Youtube session!")
			logging.exception(e)

		return 0
	
	def get_video_id(self, url: str) -> Optional[str]:
		parsed_url = urlparse(url)
		query = parse_qs(parsed_url.query)
		return query.get('v', [None])[0]

	def remove_tmp_files(self) -> None:
		for i in os.listdir(self.DOWNLOAD_DIR):
			if "yt_download_" in i:
				os.unlink(f"{self.DOWNLOAD_DIR}/{i}")

	def calculate_size_with_padding(self,
		image: Image,
		aspect_ratio_width: int,
		aspect_ratio_height: int,
		target_size: tuple=(320, 320),
		background_color: tuple=(0, 0, 0)
	) -> Image:
		aspect_ratio = aspect_ratio_width / aspect_ratio_height
		target_width, target_height = target_size

		height = target_height
		width = int(height * aspect_ratio)

		if width > target_width:
			width = target_width
			height = int(width / aspect_ratio)

		new_image = Image.new("RGB", target_size, background_color)
		image.thumbnail(target_size, Image.Resampling.LANCZOS)

		#if aspect_ratio_height > 4:
		#	image = image.resize((image.size[0], image.size[1]+new_image.size[1]))
		#	target_height += new_image.size[1] - 5

		paste_position = ((target_width - width) // 2, (target_height - height) // 2)
		new_image.paste(image, paste_position)

		return new_image

	def resize_aspect_ratio(self, image: Image, base_width: int = 320) -> Image:
		wpercent = (base_width / float(image.size[0]))
		hsize = int((float(image.size[1]) * float(wpercent)))
		img = image.resize((base_width, hsize), Image.Resampling.LANCZOS)

		return img

	def aspect_ratio(self, size: tuple) -> tuple:
		gcd = math.gcd(size[0], size[1])
		return size[0] // gcd, size[1] // gcd
	
	def crop_black_edges_pil(self, img: Image, threshold: int = 35) -> Image:
		y_nonzero, x_nonzero, _ = np.nonzero(np.array(img) > threshold)
		return img.crop((np.min(x_nonzero), np.min(y_nonzero), np.max(x_nonzero), np.max(y_nonzero)))

	def crop_center(self, image: Image, width: int, height: int) -> Image:
		img_width, img_height = image.size
		left = (img_width - width) // 2
		top = (img_height - height) // 2
		right = left + width
		bottom = top + height

		return image.crop((left, top, right, bottom))

	def download_thumbnail(self, video_id: str, timeout: int, crop_center: dict = None) -> Optional[io.BytesIO]:
		for i in ("https://img.youtube.com/vi/{VIDEO_ID}/maxresdefault.jpg",
				"https://img.youtube.com/vi/{VIDEO_ID}/hqdefault.jpg",
				"https://img.youtube.com/vi/{VIDEO_ID}/sddefault.jpg"):
			try:
				url = i.format(VIDEO_ID=video_id)
				logging.info("Youtube thumbnail url '%s'", url)
				proxies = None
				if self.proxy:
					dsn = self.proxy.get("dsn", None)
					if dsn:
						proxies = self.build_proxies(dsn)
				with requests.get(url, timeout=(timeout, timeout), proxies=proxies) as response:
					if response.status_code == 200:
						image = Image.open(io.BytesIO(response.content))
						ratio = self.aspect_ratio(image.size)
						image = self.crop_black_edges_pil(image)
						if crop_center:
							image = self.crop_center(image, width=crop_center["width"], height=crop_center["height"])
						logging.info("thumb ratio: '%s'", ratio)
						new_image = self.resize_aspect_ratio(image)
						logging.info("thumb size: '%s'", new_image.size)
						io_buf = io.BytesIO()
						new_image.save(io_buf, format='JPEG', subsampling=0, quality=95, progressive=True, optimize=False)
						logging.info("thumb size: %d kb", io_buf.getbuffer().nbytes / 1024)
						io_buf.seek(0)
						return io_buf
			except Exception as e:
				logging.error("Failed to download download thumbnail!")
				logging.exception(e)

		return None

	def download_hndlr(self, func: Callable, *args: tuple[Union[str, int, dict, tuple, bool]], **kwargs: dict[Union[str, int, dict, tuple, bool]]) -> Optional[Union[list, dict, str, io.BytesIO]]:
		ret_val = None
		max_retries = int(os.environ.get("YT_MAX_RETRIES", default=self.YT_MAX_RETRIES_DEFAULT))
		pause_secs = int(os.environ.get("YT_PAUSE_BEFORE_RETRY", default=self.YT_PAUSE_BEFORE_RETRY_DEFAULT))
		timeout = int(os.environ.get("YT_TIMEOUT", default=self.YT_TIMEOUT_DEFAULT))
		timeout_increment = int(os.environ.get("YT_TIMEOUT_INCREMENT", default=self.YT_TIMEOUT_INCREMENT_DEFAULT))
		retries = 0
		while max_retries >= retries:
			try:
				kwargs["timeout"] = timeout
				ret_val = func(*args, **kwargs)
				break
			except pytubefix.exceptions.MaxRetriesExceeded:
				# do noting, not interested
				pass
			#except http.client.IncompleteRead as e:
			except (KeyError, pytubefix.exceptions.RegexMatchError):
				raise Unavailable("Library failed")
			except urllib3.exceptions.ProxyError as e:
				logging.warning("Proxy error!")
				raise BadProxy(extract_exception_message(e.original_error))
			except (socket.timeout,
					ssl.SSLError,
					http.client.IncompleteRead,
					http.client.HTTPException,
					requests.RequestException,
					urllib.error.URLError,
					urllib.error.HTTPError) as e:
				if hasattr(e, "code") and (int(e.code) == 403 or int(e.code) == 400):
					raise Unavailable(extract_exception_message(e))
				if hasattr(e, "reason") and "Remote end closed connection without response" in str(e.reason):
					raise Unavailable(extract_exception_message(e))
				logging.warning("Youtube read timeout! Retrying in '%d' seconds ...", pause_secs)
				logging.info("Your `YT_MAX_RETRIES` values is '%d'", max_retries)
				logging.exception(extract_exception_message(e))
				if max_retries <= retries:
					#self.remove_tmp_files()
					raise TimeOut(extract_exception_message(e))
				retries += 1
				timeout += timeout_increment
				time.sleep(pause_secs)
			except (pytubefix.exceptions.VideoUnavailable, pytubefix.exceptions.VideoPrivate) as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.DownloadError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.GeoRestrictedError:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.PostProcessingError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.ExtractorError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.MaxDownloadsReached as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.UnavailableVideoError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.ThrottledDownload as e:
				raise Unavailable(extract_exception_message(e))

		return ret_val

	def yt_on_progress(self, stream: Stream, chunk: bytes, bytes_remaining: int) -> None:
		total_size = stream.filesize or stream.filesize_approx
		bytes_downloaded = total_size - bytes_remaining
		percentage_of_completion = bytes_downloaded / (total_size or 1) * 100
		if total_size == 0 or percentage_of_completion >= self._download_progress_threshold:
			msg = {
				"action": "report_download_status",
				"current": bytes_downloaded,
				"total": total_size,
				"message_id": self.job.placeholder_message_id,
				"chat_id": self.job.chat_id,
				"completed": percentage_of_completion >= 100
			}
			self.status_pipe.send(msg)
			logging.debug("[Download worker] Downloaded %d%%", percentage_of_completion)
			if total_size > 0:
				self._download_progress_threshold += 20

	def build_proxies(self, proxy_dsn: str) -> dict:
		if not proxy_dsn:
			logging.warning("Empty DSN!")
			return {}
		proxy = {}
		if proxy_dsn:
			if "https://" in proxy_dsn:
				proxy["https"] = proxy_dsn
			elif "http://" in proxy_dsn:
				proxy["http"] = proxy_dsn
			else:
				logging.warning("Proxy DSN malformed!")

		return proxy

	def build_yt(self, url: str, session: bool = True) -> YouTube:
		_default_clients["ANDROID"]["innertube_context"]["context"]["client"]["clientVersion"] = "19.08.35"
		_default_clients["ANDROID_MUSIC"] = _default_clients["ANDROID"]
		yt_opts = {"url": url, "on_progress_callback": self.yt_on_progress}
		if session:
			yt_opts["client"] = "TV_EMBED"
			yt_opts["use_oauth"] = True
			yt_opts["allow_oauth_cache"] = True
			yt_opts["token_file"] = self.YT_SESSION_FILE % self.account_index
			if not os.path.exists(yt_opts["token_file"]):
				logging.warning("YT session '%s' file is not found", yt_opts["token_file"])
				self.request_yt_auth()
				self.auth_event.wait()
				yt_auth = YtAuth(account_index=self.account_index)
				device_code = yt_auth.load_device_code()
				if device_code:
					auth_data = yt_auth.confirm_token(device_code=device_code)
					if auth_data:
						yt_auth.safe_write_session(auth_data)
				else:
					logging.error("Failed to fetch YT auth token!")
		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				yt_opts["proxies"] = self.build_proxies(proxy_dsn)
		return YouTube(**yt_opts)
	
	def build_yt_dlp(self, timeout: int = 60) -> yt_dlp.YoutubeDL:
		auth_data = {}
		with open(self.YT_SESSION_FILE % self.account_index, 'r', encoding="utf-8") as f:
			auth_data = json.loads(f.read())
		time_name = str(time.time()).replace('.', '_')
		ydl_opts = {
			'socket_timeout': timeout,
			'outtmpl': f'{self.DOWNLOAD_DIR}/yt_download_{time_name}.%(ext)s',
			'format': 'bestvideo+bestaudio/best',
			'merge_output_format': 'mp4',
			'noplaylist': True,
			'tv_auth': auth_data
		}

		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				ydl_opts["proxy"] = proxy_dsn

		return yt_dlp.YoutubeDL(ydl_opts)
	
	def _download(self, url: str, session: bool = True, thumbnail: Optional[io.BytesIO] = None, timeout: int = 60) -> list:
		raise NotImplementedError("You should to implement _download method")
	
	def _download_yt_dlp(self, url: str, thumbnail: Optional[io.BytesIO] = None, timeout: int = 60) -> list:
		raise NotImplementedError("You should to implement _download_yt_dlp method")

	def download(self, job: DownloadJob) -> list:
		self.job = job
		ret = []
		thumbnail = None
		try:
			video_id = self.get_video_id(job.url)
			# shorts custom thumb
			##vinfo = VideoInfo(local_file)
			#thumbnail = self.download_hndlr(self.download_thumbnail, video_id=yt.video_id, crop_center=vinfo.get_demensions())
			if video_id:
				thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
		except Exception as e:
			logging.error("Failed to download thumb!")
			logging.exception(e)

		try:
			self.status_pipe.send({"action": "report_download_status", "current": 0, "total": 0,
									"message_id": self.job.placeholder_message_id, "chat_id": self.job.chat_id})
			ret = self.download_hndlr(self._download, job.url, session=True, thumbnail=thumbnail)
			return ret
		except (Unavailable, TimeOut, KeyError) as e:
			logging.warning("Download failed, trying to download with yt_dlp")
			logging.exception(e)
		
		try:
			ret = self.download_hndlr(self._download_yt_dlp, job.url, thumbnail=thumbnail)
		except NotImplementedError:
			logging.info("yt_dlp is not supported for this submodule yet")
			raise Unavailable("Сontent unvailable")

		return ret



import logging

import asyncio
import hashlib

from pyrogram.enums import ParseMode
from pyrogram.errors.exceptions.bad_request_400 import MessageNotModified
from pyrogram import Client
from warp_beacon.telegram.types import ReportType

class ProgressBar(object):
	MAX_PROGRESS_RENDER_SIZE = 30_000_000

	def __init__(self, client: Client) -> None:
		self._next_threshold = 20
		self.client = client
		self.complete = False

	def is_complete(self) -> bool:
		return self.complete

	def make_progress_bar(self, current: int, total: int, length: int = 10) -> str:
		"""
		Returns string
		[████▌────] 55%
		length — amount of characters in bar
		"""
		# fraction of completed job from 0.0 to 1.0
		frac = current / total if total else 0
		# how much "filled" cells
		filled = int(frac * length)
		# part between whole cells, optional may withdraw half
		half_block = ''
		if (frac * length) - filled >= 0.5:
			half_block = '▌'  # or '▏', '▍' and etc.
		# building bar
		pbar = '█' * filled + half_block + '─' * (length - filled - len(half_block))
		percent = frac * 100
		return f"<b>[{pbar}] {round(percent)}%</b>"

	def format_size_si(self, bytes_num: int) -> str:
		units = [("TB", 1024**4), ("GB", 1024**3), ("MB", 1024**2), ("KB", 1024), ("B", 1)]
		for unit_name, unit_value in units:
			if bytes_num >= unit_value:
				value = bytes_num / unit_value
				return f"{value:.2f} {unit_name}"
		return "0 B"

	def make_emoji_progress_bar(self, percent: int, length: int = 10) -> str:
		"""
		Builds an emoji progress bar.

		Args:
			percent: int from 0 to 100
			length: total number of emoji cells

		Returns:
			String like "[🟩🟩🟩⬜️⬜️⬜️⬜️⬜️⬜️⬜️] 30%"
		"""
		filled = int(percent * length / 100 + 0.5)
		empty  = length - filled
		bar = "🟩" * filled + "⬜️" * empty
		return f"[{bar}] {percent}%"

	def _on_edit_done(self, task: asyncio.Task) -> None:
		exc = task.exception()
		if not exc:
			return
		if isinstance(exc, MessageNotModified):
			logging.warning("bad_request_400.MessageNotModified")
		else:
			logging.exception("Error in edit_message_caption", exc_info=exc)

	async def progress_callback(self, current: int, total: int, chat_id: int | str, message_id: int, operation: str, report_type: ReportType, label: str = "") -> None:
		if report_type == ReportType.PROGRESS:
			return self.render_progress_bar(current=current, total=total, chat_id=chat_id, message_id=message_id, operation=operation, label=label)
		elif report_type == ReportType.ANNOUNCE:
			return self.render_progress_announce(chat_id=chat_id, message_id=message_id, label=label)

	def render_progress_announce(self, chat_id: int | str, message_id: int, label: str) -> None:
		try:
			#await self.client.edit_message_caption(chat_id, message_id, f"{pbar} <b>{operation}</b> {label}", ParseMode.HTML)
			# we don't need to wait completion, waste of time and resources
			task = self.client.loop.create_task(
				self.client.edit_message_caption(chat_id, message_id, f"<b>{label}</b>", ParseMode.HTML)
			)
			task.add_done_callback(self._on_edit_done)
		except MessageNotModified:
			logging.warning("bad_request_400.MessageNotModified")
		except Exception as e:
			logging.warning("An error occurred while setup task to update progress bar")
			logging.exception(e)

	def render_progress_bar(self, current: int, total: int, chat_id: int | str, message_id: int, operation: str, label: str = "") -> None:
		if total <= self.MAX_PROGRESS_RENDER_SIZE:
			return
		percent = 0
		if total:
			percent = round(current * 100 / (total or 1))
		if percent >= 100:
			return
		if total == 0 or percent >= self._next_threshold:
			#pbar = self.make_progress_bar(percent, 100, 25)
			pbar = self.make_emoji_progress_bar(percent, 10)
			logging.info("[Progress bar]: Operation: %s %d%%", operation, percent)
			try:
				#await self.client.edit_message_caption(chat_id, message_id, f"{pbar} <b>{operation}</b> {label}", ParseMode.HTML)
				# we don't need to wait completion, waste of time and resources
				text = f"{pbar}\n{operation}"
				if label:
					text += f"\n<code>{label}</code>"
				if total:
					text += f" <b>{self.format_size_si(total)}</b>"
				task = self.client.loop.create_task(
					self.client.edit_message_caption(chat_id, message_id, text, ParseMode.HTML)
				)
				task.add_done_callback(self._on_edit_done)
			#except MessageNotModified:
			#	logging.warning("bad_request_400.MessageNotModified")
			except Exception as e:
				logging.warning("An error occurred while setup task to update progress bar")
				logging.exception(e)
			if total > 0:
				self._next_threshold += 20

	@staticmethod
	def make_hash(chat_id: str | int, message_id: int, algorithm: str = 'sha256') -> str:
		s = f"{chat_id}:{message_id}"
		# md5, sha1, sha256
		h = hashlib.new(algorithm, s.encode('utf-8'))
		return h.hexdigest()


import os
import signal
import asyncio
from typing import Optional, Union

import logging
import html
import uvloop

from pyrogram import Client, filters
from pyrogram.enums import ParseMode, ChatType
from pyrogram.handlers import MessageHandler, CallbackQueryHandler
from pyrogram.types import InputMediaAudio, InputMediaPhoto, InputMediaVideo, InputMediaAnimation, InlineKeyboardButton, InlineKeyboardMarkup
from pyrogram.errors import NetworkMigrate, BadRequest, MultiMediaTooLong, MessageIdInvalid

import warp_beacon
from warp_beacon.__version__ import __version__
from warp_beacon.telegram.handlers import Handlers
from warp_beacon.telegram.placeholder_message import PlaceholderMessage
from warp_beacon.storage.mongo import DBClient
from warp_beacon.storage import Storage
from warp_beacon.uploader import AsyncUploader
from warp_beacon.jobs.upload_job import UploadJob
from warp_beacon.jobs.types import JobType
from warp_beacon.jobs import Origin
from warp_beacon.telegram.utils import Utils
from warp_beacon.telegram.caption_shortener import CaptionShortner
from warp_beacon.scheduler.scheduler import IGScheduler
from warp_beacon.telegram.edit_message import EditMessage
from warp_beacon.telegram.download_status import DownloadStatus

class Bot(object):
	should_exit = None
	storage = None
	uploader = None
	downloader = None
	allow_loop = True
	client = None
	handlers = None
	placeholder = None
	scheduler = None
	me = None
	edit_message = None
	download_status = None

	def __init__(self, tg_bot_name: str, tg_token: str, tg_api_id: str, tg_api_hash: str) -> None:
		# Enable logging
		logging.basicConfig(
			format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
			level=logging.INFO
		)

		logging.getLogger("pyrogram").setLevel(logging.ERROR)
		logging.info("Starting Warp Beacon version '%s' ...", __version__)
		self.storage = Storage(DBClient())
		self.should_exit = asyncio.Event()
		workers_amount = min(32, os.cpu_count() + 4)

		uvloop.install()
		self.client = Client(
			name=tg_bot_name,
			app_version=__version__,
			bot_token=tg_token,
			api_id=tg_api_id,
			api_hash=tg_api_hash,
			workdir='/var/warp_beacon',
			workers=int(os.environ.get("TG_WORKERS_POOL_SIZE", default=workers_amount))
		)

		self.editor = EditMessage(self.client)

		self.uploader = AsyncUploader(
			storage=self.storage,
			admin_message_callback=self.send_text_to_admin,
			request_yt_auth_callback=self.request_yt_auth,
			pool_size=int(os.environ.get("UPLOAD_POOL_SIZE", default=workers_amount)),
			loop=self.client.loop
		)
		self.download_status = DownloadStatus(self.client)
		self.downloader = warp_beacon.scraper.AsyncDownloader(
			workers_count=int(os.environ.get("WORKERS_POOL_SIZE", default=workers_amount)),
			uploader=self.uploader,
			pipe_connection=self.download_status.child_conn
		)

		self.scheduler = IGScheduler(self.downloader)
		self.handlers = Handlers(self)

		self.client.add_handler(MessageHandler(self.handlers.start, filters.command("start")))
		self.client.add_handler(MessageHandler(self.handlers.help, filters.command("help")))
		self.client.add_handler(MessageHandler(self.handlers.random, filters.command("random")))
		self.client.add_handler(MessageHandler(self.handlers.handler))
		#TODO refactor to callback router
		self.client.add_handler(CallbackQueryHandler(self.handlers.simple_button_handler, filters=filters.create(lambda _, __, q: not q.data.startswith("read_more:"))))
		self.client.add_handler(CallbackQueryHandler(self.handlers.read_more_handler, filters=filters.create(lambda _, __, q: q.data.startswith("read_more:"))))

		self.placeholder = PlaceholderMessage(self)

	def __del__(self) -> None:
		self.stop()

	def start(self) -> None:
		self.client.run(self.run())

	async def run(self) -> None:
		loop = asyncio.get_running_loop()
		for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGABRT):
			loop.add_signal_handler(sig, self.should_exit.set)
		async with self.client:
			self.me = await self.client.get_me()
			self.client.me = self.me
			if self.me.is_premium:
				os.environ["TG_PREMIUM"] = "true"
			self.downloader.start()
			self.uploader.start()
			loop.add_reader(self.download_status.status_pipe.fileno(), self.download_status.on_status)
			self.scheduler.start()
			logging.info("Warp Beacon version '%s' started", __version__)
			await self.should_exit.wait()

		self.stop()
		logging.info("Warp Beacon version '%s' terminated.", __version__)

	def stop(self) -> None:
		logging.info("Warp Beacon is terminating. This may take a while ...")
		self.scheduler.stop()
		self.downloader.stop_all()
		self.uploader.stop_all()
		if self.client and self.client.is_initialized and self.client.is_connected:
			asyncio.run_coroutine_threadsafe(self.client.stop(), self.client.loop)

	async def send_text(self, chat_id: int, text: str, reply_id: int = None) -> int:
		try:
			message_reply = await self.client.send_message(
				chat_id=chat_id,
				text=text,
				parse_mode=ParseMode.HTML,
				reply_to_message_id=reply_id
			)
			return message_reply.id
		except Exception as e:
			logging.error("Failed to send text message!")
			logging.exception(e)

		return 0

	async def request_yt_auth(self) -> None:
		from warp_beacon.yt_auth import YtAuth
		acc_index, acc = self.downloader.acc_selector.get_current_for_module(Origin.YOUTUBE)
		yt_auth = YtAuth(account_index=acc_index)
		data = yt_auth.fetch_token()
		if all(data.values()):
			if yt_auth.store_device_code(data["device_code"]):
				reply_markup = InlineKeyboardMarkup(
					[
						[
							InlineKeyboardButton("✅ Done", callback_data=f"confirm_yt_auth:{acc_index}")
						]
					]
				)
				await self.send_text_to_admin(
					f"Please open {data['verification_url']} and input code <code>{data['user_code']}</code>.\n\n"
					"Please authorize with a Google account with verified age.\n"
					"This will allow you to avoid error the <b>AgeRestrictedError</b> when accessing some content.",
					account_admins=acc.get("account_admins", None),
					reply_markup=reply_markup
				)
		else:
			logging.error("Wrong YT auth dataset: '%s'", str(data))

	async def send_text_to_admin(self, text: str, account_admins: str = None, reply_markup: Optional[Union[InlineKeyboardMarkup]] = None) -> list[int]:
		try:
			admins = None
			if account_admins:
				admins = account_admins
			if not admins:
				admins = os.environ.get("TG_BOT_ADMINS_USERNAMES", None)
			if not admins:
				raise ValueError("Configuration value `TG_BOT_ADMIN_USERNAME` is empty!")
			
			msg_ids = []
			admins_array = admins.split(',')
			for adm in admins_array:
				adm = adm.strip()
				msg_opts = {"chat_id": adm, "text": text, "parse_mode": ParseMode.HTML}
				if reply_markup:
					msg_opts["reply_markup"] = reply_markup
				message_reply = await self.client.send_message(**msg_opts)
				msg_ids.append(message_reply.id)
			return msg_ids
		except Exception as e:
			logging.error("Failed to send text message to admin!")
			logging.exception(e)

		return []

	def build_signature_caption(self, job: UploadJob) -> str:
		caption = ""
		is_group = job.chat_type in (ChatType.GROUP, ChatType.SUPERGROUP)
		is_youtube = job.job_origin in (Origin.YOUTUBE, Origin.YT_SHORTS)
		if job.canonical_name:
			if is_group and not is_youtube and CaptionShortner.need_short(job.canonical_name):
				caption = f"{html.escape(CaptionShortner.smart_truncate_html(job.canonical_name))} ..."
				job.short_text = True
			elif is_group:
				# short enough
				caption = html.escape(job.canonical_name)
			# Captions only for YouTube in groups; empty otherwise.
			# Might be too long for Telegram, so we skip them.
			if is_youtube and is_group:
				caption = f"<b>{html.escape(job.canonical_name)}</b>"
		if is_group:
			if job.canonical_name:
				caption += "\n—\n"
			if job.message_leftover:
				caption += job.message_leftover
				caption += "\n—\n"
			if job.source_username:
				if job.source_username.startswith("<a href="):
					caption += f"Requested by <b>{job.source_username}</b>"
				else:
					caption += f"Requested by <b>@{job.source_username}</b>"
			if job.source_username and job.url:
				caption += " | "
			if job.url:
				caption += f'<a href="{job.url}">source link</a>'

		return caption

	def build_tg_args(self, job: UploadJob) -> dict:
		args = {}
		if job.media_type == JobType.VIDEO:
			if job.tg_file_id:
				if job.placeholder_message_id:
					args["media"] = InputMediaVideo(
						media=job.tg_file_id.replace(":video", ''),
						supports_streaming=True,
						caption=self.build_signature_caption(job)
					)
				else:
					args["video"] = job.tg_file_id.replace(":video", '')
					args["caption"] = self.build_signature_caption(job)
			else:
				if job.placeholder_message_id:
					args["media"] = InputMediaVideo(
						media=job.local_media_path,
						supports_streaming=True,
						width=job.media_info["width"],
						height=job.media_info["height"],
						duration=round(job.media_info["duration"]),
						thumb=job.media_info["thumb"],
						caption=self.build_signature_caption(job)
					)
				else:
					args["video"] = job.local_media_path
					args["supports_streaming"] = True
					args["width"] = job.media_info["width"]
					args["height"] = job.media_info["height"]
					args["duration"] = round(job.media_info["duration"])
					args["thumb"] = job.media_info["thumb"]
					args["caption"] = self.build_signature_caption(job)

				args["file_name"] = os.path.basename(job.local_media_path)
		elif job.media_type == JobType.IMAGE:
			if job.tg_file_id:
				if job.placeholder_message_id:
					args["media"] = InputMediaPhoto(
						media=job.tg_file_id.replace(":image", ''),
						caption=self.build_signature_caption(job)
					)
				else:
					args["photo"] = job.tg_file_id.replace(":image", '')
					args["caption"] = self.build_signature_caption(job)
			else:
				if job.placeholder_message_id:
					args["media"] = InputMediaPhoto(
						media=job.local_media_path,
						caption=self.build_signature_caption(job)
					)
				else:
					args["photo"] = job.local_media_path
					args["caption"] = self.build_signature_caption(job)

				args["file_name"] = os.path.basename(job.local_media_path)
		elif job.media_type == JobType.AUDIO:
			if job.tg_file_id:
				if job.placeholder_message_id:
					args["media"] = InputMediaAudio(
						media=job.tg_file_id.replace(":audio", ''),
						caption=self.build_signature_caption(job)
					)
				else:
					args["audio"] = job.tg_file_id.replace(":audio", '')
			else:
				if job.placeholder_message_id:
					args["media"] = InputMediaAudio(
						media=job.local_media_path,
						performer=job.media_info["performer"],
						thumb=job.media_info["thumb"],
						duration=round(job.media_info["duration"]),
						title=job.canonical_name,
						caption=self.build_signature_caption(job)
					)
				else:
					args["audio"] = job.local_media_path
					args["performer"] = job.media_info["performer"]
					args["thumb"] = job.media_info["thumb"]
					args["duration"] = round(job.media_info["duration"])
					args["title"] = job.canonical_name
					args["caption"] = self.build_signature_caption(job)

				args["file_name"] = os.path.basename(job.local_media_path)
		elif job.media_type == JobType.ANIMATION:
			if job.tg_file_id:
				if job.placeholder_message_id:
					args["media"] = InputMediaAnimation(
						media=job.tg_file_id.replace(":animation", ''),
						caption=self.build_signature_caption(job)
					)
				else:
					args["animation"] = job.tg_file_id.replace(":animation", '')
					args["caption"] = self.build_signature_caption(job)
			else:
				if job.placeholder_message_id:
					args["media"] = InputMediaAnimation(
						media=job.local_media_path,
						thumb=job.media_info["thumb"],
						duration=round(job.media_info["duration"]),
						width=job.media_info["width"],
						height=job.media_info["height"],
						caption=self.build_signature_caption(job)
					)
				else:
					args["animation"] = job.local_media_path
					args["width"] = job.media_info["width"]
					args["height"] = job.media_info["height"]
					args["duration"] = round(job.media_info["duration"])
					args["thumb"] = job.media_info["thumb"]
					args["caption"] = self.build_signature_caption(job)

				args["file_name"] = os.path.basename(job.local_media_path)
		elif job.media_type == JobType.COLLECTION:
			if job.tg_file_id:
				args["media"] = []
				for chunk in Utils.chunker(job.tg_file_id.split(','), 10):
					tg_chunk = []
					for i in chunk:
						tg_id, mtype = i.split(':')
						ctype = JobType[mtype.upper()]
						ptr = None
						if ctype == JobType.VIDEO:
							ptr = InputMediaVideo(media=tg_id, caption=self.build_signature_caption(job))
						elif ctype == JobType.IMAGE:
							ptr = InputMediaPhoto(media=tg_id, caption=self.build_signature_caption(job))
						elif ctype == JobType.ANIMATION:
							ptr = InputMediaAnimation(media=tg_id, caption=self.build_signature_caption(job))
						tg_chunk.append(ptr)

					args["media"].append(tg_chunk)
			else:
				mediafs = []
				for chunk in job.media_collection:
					tg_chunk = []
					for j in chunk:
						if j.media_type == JobType.VIDEO:
							vid = InputMediaVideo(
								media=j.local_media_path,
								supports_streaming=True,
								width=j.media_info["width"],
								height=j.media_info["height"],
								duration=round(j.media_info["duration"]),
								thumb=j.media_info["thumb"],
								caption=self.build_signature_caption(job)
							)
							tg_chunk.append(vid)
						elif j.media_type == JobType.IMAGE:
							photo = InputMediaPhoto(
								media=j.local_media_path,
								caption=self.build_signature_caption(job)
							)
							tg_chunk.append(photo)
					mediafs.append(tg_chunk)
				args["media"] = mediafs

		args["chat_id"] = job.chat_id

		# common args
		if job.placeholder_message_id and job.media_type is not JobType.COLLECTION:
			args["message_id"] = job.placeholder_message_id
		else:
			args["disable_notification"] = True
			args["reply_to_message_id"] = job.message_id

		if job.media_type is not JobType.COLLECTION:
			render_donates = os.environ.get("ENABLE_DONATES", None) == "true"
			keyboard_buttons = [[]]
			# short_text is True only for groups
			if job.short_text:
				keyboard_buttons[0].append(InlineKeyboardButton("📖 Read more", callback_data=f"read_more:{job.job_origin.value}:{job.uniq_id}"))
			if render_donates:
				keyboard_buttons[0].append(InlineKeyboardButton("❤ Donate", url=os.environ.get("DONATE_LINK", "https://pay.cryptocloud.plus/pos/W5BMtNQt5bJFoW2E")))

			if keyboard_buttons[0]:  #job.short_text or render_donates:
				args["reply_markup"] = InlineKeyboardMarkup(keyboard_buttons)

		return args

	async def upload_job(self, job: UploadJob) -> list[str]:
		tg_file_ids = []
		try:
			retry_amount = 0
			max_retries = int(os.environ.get("TG_MAX_RETRIES", default=5))
			while not retry_amount >= max_retries:
				try:
					reply_message = None
					if job.media_type in (JobType.VIDEO, JobType.IMAGE, JobType.AUDIO, JobType.ANIMATION):
						if job.media_type in (JobType.VIDEO, JobType.AUDIO):
							await Utils.ensure_me_loaded(self.client)
						if job.placeholder_message_id:
							try:
								reply_message = await self.editor.edit(**self.build_tg_args(job))
							except MessageIdInvalid:
								logging.warning("Placeholder message not found. Looks like placeholder message was deleted by administrator.")
								job.placeholder_message_id = None
								continue
						else:
							send_funcs = {
								JobType.VIDEO: self.client.send_video,
								JobType.IMAGE: self.client.send_photo,
								JobType.AUDIO: self.client.send_audio,
								JobType.ANIMATION: self.client.send_animation
							}
							try:
								reply_message = await send_funcs[job.media_type](**self.build_tg_args(job))
							except ValueError as e:
								err_text = str(e)
								if "Expected" in err_text:
									logging.warning("Expectations exceeded reality.")
									logging.warning(err_text)
									expectation, reality = Utils.parse_expected_patronum_error(err_text)
									job_args = self.build_tg_args(job)
									job_args[reality.value.lower()] = job_args.pop(expectation.value.lower())
									reply_message = await send_funcs[reality](**job_args)

						tg_file_id = Utils.extract_file_id(reply_message)
						tg_file_ids.append(tg_file_id)
						job.tg_file_id = tg_file_id
						logging.info("Uploaded media file with type '%s' tg_file_id is '%s'", job.media_type.value, job.tg_file_id)
					elif job.media_type == JobType.COLLECTION:
						col_job_args = self.build_tg_args(job)
						sent_messages = []
						snd_grp_options = {"chat_id": job.chat_id, "reply_to_message_id": job.message_id}
						for i, media_chunk in enumerate(col_job_args["media"]):
							snd_grp_options["media"] = media_chunk
							messages = await self.client.send_media_group(**snd_grp_options)
							sent_messages += messages
							if job.media_collection:
								for j, _ in enumerate(media_chunk):
									tg_file_id = Utils.extract_file_id(messages[j])
									if tg_file_id:
										job.media_collection[i][j].tg_file_id = tg_file_id
							if i == 0 and job.placeholder_message_id:
								await self.placeholder.remove(job.chat_id, job.placeholder_message_id)
						for msg in sent_messages:
							if msg.video:
								tg_file_ids.append(msg.video.file_id + ':video')
							elif msg.photo:
								tg_file_ids.append(msg.photo.file_id + ':image')
					logging.info("Uploaded to Telegram")
					if job.chat_type not in (ChatType.GROUP, ChatType.SUPERGROUP):
						if job.canonical_name:
							await self.send_text(chat_id=job.user_id, text=job.canonical_name)
					break
				except MultiMediaTooLong as e:
					logging.error("Failed to upload due telegram limitations :(")
					logging.exception(e)
					await self.placeholder.remove(job.chat_id, job.placeholder_message_id)
					await self.send_text(job.chat_id, e.MESSAGE, job.message_id)
					break
				except (NetworkMigrate, BadRequest, AttributeError) as e:
					import traceback
					logging.error("Network error. Check you Internet connection.")
					logging.exception(e)

					if retry_amount+1 >= max_retries:
						msg = ""
						if hasattr(e, "MESSAGE") and e.MESSAGE:
							msg = f"Telegram error: {str(e.MESSAGE)}"
						else:
							msg = (f"Unknown Telegram error. Known information:\n<pre language=\"python\">\n{traceback.format_exc().strip()}</pre>"
									"\nPlease <a href=\"https://github.com/sb0y/warp_beacon/issues\">create issue</a> with this info and service logs.")
						await self.placeholder.remove(job.chat_id, job.placeholder_message_id)
						await self.send_text(job.chat_id, msg, job.message_id)
						break
				retry_amount += 1
		except Exception as e:
			logging.error("Error occurred!")
			logging.exception(e)
		finally:
			job.remove_files()

		if job.chat_type in (ChatType.GROUP, ChatType.SUPERGROUP):
			try:
				await self.client.delete_messages(job.chat_id, (job.message_id,))
			except Exception as e:
				logging.warning("Failed to delete source message. Check bot permissions in Telegram chat settings.")
				logging.exception(e)

		return tg_file_ids